#include <u.h>
#include <libc.h>

/* UUID generator - made by Nils O. Selåsdal <NOS@Utel.no>
   Based on descriptions in draft-mealling-uuid-urn-05.txt 

   0               1               2               3
   0 1 2 3 4 5 6 7 0 1 2 3 4 5 6 7 0 1 2 3 4 5 6 7 0 1 2 3 4 5 6 7
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                          time_low                             |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |       time_mid                |         time_hi_and_version   |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |clk_seq_hi_res |  clk_seq_low  |         node (0-1)            |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                         node (2-5)                            |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
*/

/*Difference between 15 Oct 1582 and Jan 1 1970 
  in 100 nanoseconds
*/
#define TIME_DIFF 122192928000000000L 

/* The various versions */

enum {
	V_TIME = 0x1,
	V_DEC  = 0x2,
	V_MD5  = 0x3,
	V_RAND = 0x4,
	V_SHA1 = 0x5,
	VARIANT = 0x4
};
char *reverse(char *s)
{
	int i,j;
	char c;

	for(i = 0, j = strlen(s) - 1; i < j; i++,j--) {
		c = s[i];
		s[i] = s[j];
		s[j] = c;
	}
	
	return s;	
}

uvlong get_currtime(void)
{
	uvlong unixnsec = nsec();
	
	return (unixnsec/100) + TIME_DIFF;
}

ushort extract_clockseq(char *str,uvlong currtime)
{
	char *fields[2];
	uvlong saved_time;
	ushort saved_clockseq;

	if(getfields(str,fields,2,0,":") == 2) {
		saved_clockseq = atoi(fields[0]);
		saved_time = strtoull(fields[1],nil,10);
		
		if(saved_time >= currtime)
			 saved_clockseq++;
		
		return saved_clockseq;
	}
	
	return rand();
}

/* get the clockseq, if it can be read from the saved state,
   check if saved time is equal or larger than current time.
   If so, increment clockseq.
   If no savefile exists, generate a random clockseq
   Then save the clockseq and current time
*/
ushort get_clockseq(char *ether_addr,uvlong currtime)
{
	ushort saved_clockseq;
	int fd,cnt;
	char name[255];
	char err[ERRMAX],saved[256];

	sprint(name,"/tmp/uuid.%s",ether_addr);

	do {
		fd = open(name,ORDWR);
		errstr(err,sizeof(err));
	} while(fd < 0 && strcmp(err,"exclusive lock") == 0);

	saved_clockseq =  rand();

	if(fd < 0) {
		
		fd = create(name,ORDWR|OEXCL,DMTMP|DMEXCL|0660);
		if(fd < 0) {
			fprint(2,"Error, can't create state file %s : %r\n",name);
			exits("error");
		}


	} else if((cnt = read(fd,saved,sizeof(saved)-1)) > 0 ) {
		saved[cnt] = 0;
		saved_clockseq = extract_clockseq(saved,currtime);
	}		
	

	if(seek(fd,0,0) == 0) 
		fprint(fd,"%ud:%llud\n",(uint)saved_clockseq,currtime);
	else 
		fprint(2,"Warning, could not save state to %s : %r\n",name);

	close(fd);
	
	return saved_clockseq;
}

char *get_etheraddr(int inr)
{
	static char buf[16];
	char iface[32];
	int fd,cnt;
	
	sprint(iface,"/net/ether%d/addr",inr);
	fd = open(iface,OREAD);
	if(fd <0)
		return nil;
	cnt = read(fd,buf,12);
	close(fd);
	if(cnt != 12)
		return nil;
	buf[12] = 0;
	
	return buf;

}

void main(int argc,char *argv[])
{
	char *etheraddr;
	uvlong currtime;
	ushort clockseq;
	
	uint time_low,time_mid,time_hi_and_version;
	uint clk_seq_hi_res,clk_seq_low;

	srand((getpid()<<8) | time(0));
	
	etheraddr = get_etheraddr(0);
	if(etheraddr == nil) {
		fprint(2,"Error, failed to get ether address\n");
		exits("error");
	}

	currtime = get_currtime();
	clockseq = get_clockseq(etheraddr,currtime);

	time_low = currtime & 0xffffffff;
	time_mid = currtime >>32;
	time_mid &= 0xffff;
	time_hi_and_version = currtime >> 48;
	time_hi_and_version &= 0x0fff|V_TIME; 
	clk_seq_hi_res = clockseq >>8;
	clk_seq_hi_res &= 0x3f | VARIANT;
	clk_seq_low = clockseq &0xff;

	print("%08ux-%04ux-%04ux-%04ux-%s\n",time_low,time_mid,
		time_hi_and_version,(clk_seq_hi_res<<8) | clk_seq_low,
		reverse(etheraddr));

	exits(nil);
}
